/**
 * @name:    Game of Life
 *
 * @brief    Game of Life Application
 *
 * @author: Tomasz Gebarowski <gebarowski@gmail.com>
 * @date:   Wednesday 03 Nov 2008, 8:59 AM
 */
#include <jcb_common.h>
#include <jcb_board.h>
#include <jcb_sub_board.h>
#include <pp_timer.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <mpi.h>
#include <unistd.h>

/**
 * @brief  Accuracy
 */
#define JCB_ACCURACY      0.005

/**
 * @brief  Board X and Y size
 */
#define JCB_BOARD_SIZE    10

/** -------------------------------------------------------------------------*/
/** ------------------------------- Typedefs  -------------------------------*/
/** -------------------------------------------------------------------------*/

typedef enum
{
    TAG_FIRST_ROW = 0,
    TAG_LAST_ROW  = 1,
    TAG_FINAL_RESULT = 2
} row_tag_t;


/**
 * @param Parse command line parameters
 */
bool static 
__parse_cmd_line_args(int argc,
                      char *argv[],
                      unsigned int *size,
                      unsigned int *steps)
{
    bool ret_val = TRUE;

    if (argc == 3)
    {
        unsigned int tmp_size  = atoi(argv[1]);
        unsigned int tmp_steps = atoi(argv[2]);

        if (tmp_size < 2 || tmp_size > 1000)
        {
            jcb_message("Game Board size must be inside interval <2;1000>\n");
            ret_val = FALSE;
        }
        else if (tmp_steps <= 0)
        {
            jcb_message("Invalid number of iteration steps: %d\n", tmp_steps);
            ret_val = FALSE;
        }

        if (ret_val)
        {
            if (size != NULL)
                *size = tmp_size;

            if (steps != NULL)
                *steps = tmp_steps;
        }
    }
    else
    {
        jcb_message("Usage %s [MATRIX_SIZE] [STEPS]\n", argv[0]);
        ret_val = FALSE;
    }

    return ret_val;
}

/** -------------------------------------------------------------------------*/
/** ----------------------------------- Main --------------------------------*/
/** -------------------------------------------------------------------------*/

int main(int argc, char *argv[])
{
    int rank, size;

    /** Matrix Dimensions */
    unsigned int mat_width  = JCB_BOARD_SIZE;
    unsigned int mat_height = JCB_BOARD_SIZE;
    /** Max Sub Matrix Height */
    unsigned int max_height = 0;
    /** Last Sub Matrix Height */
    unsigned int last_height = 0;
    /** Steps */
    unsigned int steps      = 0;

    /** Execution timer */
    pp_timer_t *p_timer = pp_timer_create();

    /** Current Sub Board instance */
    jcb_sub_board_t *p_sub_board =  NULL;

    /** Sub Board Height */
    unsigned int sub_height = 0;

    /** Average Number of rows of Game Board per process */
    unsigned int n_rows_per_proc = 0;

    MPI_Init( &argc, &argv );
    MPI_Comm_rank( MPI_COMM_WORLD, &rank );
    MPI_Comm_size( MPI_COMM_WORLD, &size );

    /** Parse Command Line parameters only in Root Process */
    if (rank == 0)
    {
        if ( !__parse_cmd_line_args(argc,
                                    argv,
                                    &mat_width,
                                    &steps))
        {
            MPI_Finalize();
            return 66;
        }
    }

    /** Broadcast parameters to all processes */
    MPI_Bcast(&mat_width, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&steps,1, MPI_INT, 0, MPI_COMM_WORLD);
    
    mat_height = mat_width;

    /** Calculate average number of rows per process*/
    n_rows_per_proc = mat_height / size;

    if ( size > mat_height )
    {
        jcb_message("Number of processes cannot be greater then rows number\n");
        MPI_Finalize();
        return - 1;
    }

    sub_height  = n_rows_per_proc;

    /** For the last process calculate remaining number of rows */
    if (rank == size - 1  &&
        size > 1)
    {
        sub_height  = mat_height - (size - 1) * n_rows_per_proc;
        last_height = sub_height;
    }

    MPI_Bcast(&last_height,1, MPI_INT, size - 1, MPI_COMM_WORLD);

 
    /** Calculate Max Height */
    max_height = MAX(last_height,sub_height);
   
    /** Broadcast last sub matrix height */
    MPI_Bcast(&max_height,1, MPI_INT, size - 1, MPI_COMM_WORLD);


    /** Create Sub Board */
    p_sub_board = jcb_sub_board_create(sub_height, 
                                       mat_width);

    /** Init with Top Boundry */
    if (rank == 0 )
    {
        if (size == 1)
        {
            jcb_sub_board_init(p_sub_board, JCB_TOP_BOUNDARY | JCB_BOTTOM_BOUNDARY);
        }
        else
        {
            jcb_sub_board_init(p_sub_board, JCB_TOP_BOUNDARY);
        }
    }
    /** Init with Bottom Boundry */
    else if (rank == size - 1 )
    {
         jcb_sub_board_init(p_sub_board, JCB_BOTTOM_BOUNDARY);
    }
    /** Init as LEFT & RIGHT boundary only */
    else
    {
         jcb_sub_board_init(p_sub_board, JCB_OTHER_BOUNDARY);
    }

    if (p_sub_board != NULL)
    {
        jcb_board_t  *p_jcb_board = NULL;

        MPI_Request request_arr[2];
        MPI_Status  status_arr[2];

        double sub_avg = 0; /* Sub Matrix Average    */
        double avg     = 0; /* Current Total Average */
        double old_avg = 0; /* Old total average     */
        double mean_error= 1111; /* Mean error        */

        double first_row[mat_width];
        double last_row[mat_width];
        unsigned int step = 0;

        if (rank == 0)
        {
            jcb_message("Matrix [%d x %d] Steps: %d\n",
                         mat_width, mat_height,steps);
            jcb_message("Processing...\n");
            p_jcb_board = jcb_board_create(mat_height,mat_width);
        }

        /** Start timer for Root Thread */
        if (rank == 0)
        {
            if (p_timer != NULL)
                pp_timer_start(p_timer);
        }

        /** Perform action until a given accuracy is achieved 
          * or until a given number of iterations is exceeded*/
        while(mean_error > JCB_ACCURACY &&
              step < steps)
        {
            if (rank == 0)
            {
                /** Remember Old Avg */
                old_avg = avg;
            }

            /** ---------- Exchange Neighbours ----------- */
            MPI_Irecv( last_row, 
                       mat_width,
                       MPI_DOUBLE,
                       jcb_prev_rank(rank,size),
                       TAG_LAST_ROW,
                       MPI_COMM_WORLD,
                       &request_arr[0] );

            MPI_Irecv( first_row, 
                       mat_width,
                       MPI_DOUBLE,
                       jcb_next_rank(rank,size),
                       TAG_FIRST_ROW,
                       MPI_COMM_WORLD,
                       &request_arr[1] );

            /** Send first row to rank-1 processs */
            MPI_Send( p_sub_board->p_arr, 
                      mat_width,
                      MPI_DOUBLE,
                      jcb_prev_rank(rank,size),
                      TAG_FIRST_ROW,
                      MPI_COMM_WORLD );

            /** Send last row to  rank+1 process  */
            MPI_Send( jcb_last_row_ptr(p_sub_board->p_arr, sub_height, mat_width ),
                      mat_width,
                      MPI_DOUBLE,
                      jcb_next_rank(rank,size),
                      TAG_LAST_ROW,
                      MPI_COMM_WORLD );

            /** Wait for both last and first raw neighbours */
            MPI_Waitall(2, request_arr, status_arr );
    
            /** Create temporary sub-game boards */
            jcb_sub_board_create_tmp_src(p_sub_board, first_row, last_row, mat_width );
       
            /** Run single iteration step*/
            jcb_sub_board_run_step(p_sub_board);
    


            /** Compute Number of Alive cells in sub boards */
            jcb_sub_board_avg(p_sub_board,&sub_avg);

            /** Sum alive cells from sub boards */
            MPI_Reduce(&sub_avg, &avg, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);

            /* Calculate Average */
            
            if (rank == 0)
            {
                avg /= 1.0f * size;
                mean_error = ABS(old_avg - avg);
            
                if (step >= steps - 1 &&
                    mean_error > JCB_ACCURACY)
                {
                    jcb_message("After %d steps accuracy was not achieved (mean_error:%f, accuracy:%f)\n",
                                 step + 1, mean_error,JCB_ACCURACY);
                }
            }

            /** Broadcast Mean error to all other processes */
            MPI_Bcast(&mean_error, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);

            step++;
        }

        if (rank == 0)
        {
            jcb_board_buffer_create(p_jcb_board, max_height * size, mat_width );
        }

        /** Gather results */
        MPI_Gather(p_sub_board->p_arr, max_height * mat_width, MPI_DOUBLE,
                  (p_jcb_board != NULL) ? p_jcb_board->p_tmp_buf : NULL,
                  max_height * mat_width, MPI_DOUBLE, 0, MPI_COMM_WORLD);

        if (rank == 0)
        {
            
            if (p_timer != NULL)
            {
                pp_timer_stop(p_timer);
                jcb_message("Steps performed: %d\n", step);
                jcb_message("Measured time: %d msecs\n", pp_timer_measured_time(p_timer));
            }

            jcb_board_convert(p_jcb_board, sub_height, max_height, last_height, size);

            //jcb_board_print(p_jcb_board);
            jcb_board_write_gnuplot(p_jcb_board);
            printf("Mean error:%f\n",mean_error);
            jcb_board_dispose(p_jcb_board);
        }

        /** Dispose Sub Game Board */
        jcb_sub_board_dispose(p_sub_board);
    }
    else
    {
        jcb_message("Cannot create jcb_sub_board_t object");
    }               

    MPI_Finalize();
    return 0;
}
