/**
 * @name:    Game of Life
 *
 * @brief    Game of Life board implementation
 *
 * @author: Tomasz Gebarowski <gebarowski@gmail.com>
 * @date:   Wednesday 03 Nov 2008, 8:59 AM
 */
#include <jcb_common.h>
#include <jcb_board.h>
#include <jcb_sub_board.h>
#include <limits.h>
#include <string.h>



/** -------------------------------------------------------------------------*/
/** ------------------------ Function Implementation ------------------------*/
/** -------------------------------------------------------------------------*/

/**
 * @brief Allocates memory for Game Board
 */
jcb_board_t *
jcb_board_create(unsigned int rows,
                 unsigned int cols)
{
    jcb_board_t * p_board = malloc(sizeof(jcb_board_t));

    if (p_board != NULL)
    {
        p_board->rows = rows;
        p_board->cols = cols;
        p_board->p_arr = calloc(rows * cols, sizeof(double));

        if (p_board->p_arr == NULL)
        {
            jcb_message("Cannot allocate memory for Game Board");
            free(p_board);
            return NULL;
        }
    }
    return p_board;
}

/**
 * @brief Allocates memory for tmp buffer
 *
 * @param self Pointer to matrix instance
 *
 * @param rows
 * @param cols
 *
 * @return State of method invocation
 */
jcb_status_t
jcb_board_buffer_create(jcb_board_t *self,
                        unsigned int rows, 
                        unsigned int cols)
{
    jcb_status_t ret_code = JCB_OK;

    if (self != NULL)
    {
        self->p_tmp_buf = calloc(rows * cols, sizeof(double));

        if (self->p_tmp_buf == NULL)
        {
            jcb_message("Cannot allocate memory for Game Board");
            ret_code = JCB_ENOMEM;
        }
    }
    else
    {
        jcb_message("Invalid pointer self(%p)", self);
        ret_code = JCB_EINVPTR;
    }

    return ret_code;
}

/**
 * @brief Converts result from tmp buffer into square Matrix
 *        By removing rubbish which was created during MPI_Gather
 *
 */
jcb_status_t
jcb_board_convert(jcb_board_t *self,
                  unsigned int std_height,
                  unsigned int max_height,
                  unsigned int last_height,
                  unsigned int parts)
{
    jcb_status_t ret_code = JCB_OK;

    if (self != NULL)
    {
        unsigned int pos = 0;
        unsigned int cols = self->cols;

        for (pos = 0; pos < parts; pos++)
        {
            if (pos != parts - 1)
            {
                memcpy(&self->p_arr[pos * std_height * cols],
                       &self->p_tmp_buf[pos * max_height * cols],
                       sizeof(double)* std_height * cols);
            }
            else
            {
                memcpy(&self->p_arr[pos * std_height * cols],
                       &self->p_tmp_buf[pos * max_height * cols],
                       sizeof(double) * last_height * cols);
            }
        }
    }
    else
    {
        jcb_message("Invalid pointer self(%p)", self);
        ret_code = JCB_EINVPTR;
    }

    return ret_code;
}




/**
 * @brief Print Game Board
 */
void
jcb_board_print(jcb_board_t *self)
{
    if (self != NULL)
    {
        if (self->p_arr != NULL)
        {
            int i = 0;
            int j = 0;

            for( i = 0; i < self->rows; i++)
            {
                for( j = 0; j < self->cols; j++)
                {
                    double value = self->p_arr[self->cols * i +j];
                    printf("%0.5f ", value);
                }
                printf("\n");
            }
        }
    }
}

/**
 * @brief Writes final result to gnuplot data file
 */
jcb_status_t
jcb_board_write_gnuplot(jcb_board_t *self)
{
    jcb_status_t ret_code = JCB_OK;

    if (self != NULL)
    {
        int i,j;
        FILE* fd;

        fd = fopen(GNUPLOT_FILE_PATH, "w");

        if (fd > 0)
        {
            for (i = 0; i < self->rows; i++)
            {
                for (j = 0; j < self->cols; j++)
                {
                    double value = self->p_arr[self->cols * i +j];
                    fprintf(fd, "%f %f %11.4f\n", 1.0/(self->cols-1)*j, 1.0/(self->rows-1)*i, value);
                }
             fprintf(fd, "\n");     
            }
            fclose(fd);
        }
        else
        {
            jcb_message("Cannot create file for gnuplot data");
            ret_code = JCB_ERROR;
        }
    }
    else
    {
        jcb_message("Invalid pointer self(%p)",self);
        ret_code = JCB_EINVPTR;
    }

    return ret_code;
}



/**
 * @brief Disposes object 
 */
void
jcb_board_dispose(jcb_board_t *self)
{
    if (self != NULL)
    {
        if (self->p_arr != NULL)
        {
            free(self->p_arr);
        }

        if (self->p_tmp_buf != NULL)
        {
            free(self->p_tmp_buf);
        }

        free(self);
    }
}
