/**
 * @name:    Game of Life
 *
 * @brief    Game of Life sub_board implementation
 *
 * @author: Tomasz Gebarowski <gebarowski@gmail.com>
 * @date:   Wednesday 03 Nov 2008, 8:59 AM
 */
#include <jcb_common.h>
#include <jcb_sub_board.h>
#include <math.h>
#include <limits.h>
#include <string.h>


/** -------------------------------------------------------------------------*/
/** -------------------------- Private Typededs -----------------------------*/
/** -------------------------------------------------------------------------*/

/**
 * @brief Defines realtive position of a neighbour 
 *        with respect to investigated cell
 */
typedef struct
{
    short dx;
    short dy;
} jcb_neighbour_t;

/** -------------------------------------------------------------------------*/
/** -------------------------- Predefinitions  ------------------------------*/
/** -------------------------------------------------------------------------*/

static jcb_status_t
__jcb_sub_board_get_cell(jcb_sub_board_t *self,
                       unsigned int i,
                       unsigned int j,
                       double* value);


static jcb_status_t
__jcb_sub_board_process_cell(jcb_sub_board_t *self,
                             unsigned int i,  
                             unsigned int j,    
                             double *cell);

/** -------------------------------------------------------------------------*/
/** -------------------------- Global Variables -----------------------------*/
/** -------------------------------------------------------------------------*/


/**
 * @brief Definition of all possible neighbors of a cell
 */
static jcb_neighbour_t g_nbs[4] = {         {-1,0}, 
                                    {0,-1},        {0,1},
                                             {1,0}
                                  };
/** -------------------------------------------------------------------------*/
/** ------------------------ Function Implementation ------------------------*/
/** -------------------------------------------------------------------------*/

/**
 * @brief Allocates memory for Game Board
 */
jcb_sub_board_t *
jcb_sub_board_create(unsigned int rows,
                     unsigned int cols)
{
    jcb_sub_board_t * p_sub_board = malloc(sizeof(jcb_sub_board_t));

    if (p_sub_board != NULL)
    {
        p_sub_board->rows = rows;
        p_sub_board->cols = cols;
        p_sub_board->p_arr = calloc(rows * cols, sizeof(double));
        p_sub_board->p_tmp_src  = NULL;
        p_sub_board->p_top_boundary = NULL;
        p_sub_board->p_bot_boundary = NULL;


        if (p_sub_board->p_arr == NULL)
        {
            jcb_message("Cannot allocate memory for Game Board");
            free(p_sub_board);
            return NULL;
        }
    }
    return p_sub_board;
}

/**
 * @brief Initializes sub_board with 
 *        a proper boundary type
 */
jcb_status_t
jcb_sub_board_init(jcb_sub_board_t *self,
                   unsigned char boundary_type)
{
    jcb_status_t ret_code = JCB_OK;

    if  (self != NULL)
    {
        unsigned int i = 0;

        self->boundary_type = boundary_type;
        self->p_top_boundary = NULL;
        self->p_bot_boundary = NULL;

        if (boundary_type & JCB_TOP_BOUNDARY)
        {
            self->p_top_boundary = malloc(sizeof(double) * self->cols);

            if (self->p_top_boundary != NULL)
            {
                for (i = 0; i < self->cols; i++)
                {
                    self->p_top_boundary[i] = sin(M_PI * i / self->cols);
                }
            }
            else
            {
                jcb_message("Cannot allocate memory for submatrix boundary\n");
                ret_code = JCB_ENOMEM;
            }
        }
        if (boundary_type & JCB_BOTTOM_BOUNDARY)
        {
            self->p_bot_boundary = malloc(sizeof(double) * self->cols);

            if (self->p_bot_boundary != NULL)
            {
                for (i = 0; i < self->cols; i++)
                {
                    self->p_bot_boundary[i] = sin(M_PI * i / self->cols) * exp(-M_PI);
                }
            }
            else
            {
                jcb_message("Cannot allocate memory for submatrix boundary");
                ret_code = JCB_ENOMEM;
            }
        }
    }
    else
    {
        jcb_message("Invalid pointers self(%p)\n", self);
        ret_code = JCB_EINVPTR;
    }

    return ret_code;
}


/**
 * @brief Creates Source Sub Board by concatenating
 *        first_row with sub_board and p_last_row
 */
jcb_status_t
jcb_sub_board_create_tmp_src(jcb_sub_board_t *self, 
                             double *p_first_row,
                             double *p_last_row,
                             unsigned int  row_size)
{
    jcb_status_t ret_code = JCB_OK;

    if  (self != NULL &&
         p_first_row != NULL &&
         p_last_row  != NULL)
    {
        if (self->cols == row_size)
        {
            unsigned int arr_size = self->rows * self->cols;

            if (self->p_tmp_src != NULL)
                free(self->p_tmp_src);

            self->p_tmp_src = malloc(sizeof(double) * (row_size * 2 + arr_size));

            if (self->p_tmp_src != NULL)
            {
                /** Append first row */
                memcpy(self->p_tmp_src, p_last_row, row_size * sizeof(double));
                /** Append current cells */
                memcpy(&(self->p_tmp_src[row_size]), self->p_arr, arr_size * sizeof(double));
                /** Append second row */
                memcpy(&(self->p_tmp_src[arr_size + row_size]), p_first_row, row_size * sizeof(double));
            }
            else
            {
                jcb_message("Not enough memory\n");
                ret_code = JCB_ENOMEM;
            }
        }
        else
        {
            jcb_message("Wrong size of a row\n");
            ret_code = JCB_ERROR;
        }
    }
    else
    {
        jcb_message("Invalid pointers self(%p) p_first_row(%p) p_last_row(%p)\n",
                     self, p_first_row, p_last_row);
        ret_code = JCB_EINVPTR;
    }

    return ret_code;
}

/**
 * @brief Calculates Average of all values inside sub matrix
 *        
 */
jcb_status_t
jcb_sub_board_avg(jcb_sub_board_t *self,
                  double          *value)
{
    jcb_status_t ret_code = JCB_EINVPTR;

    if (self  != NULL &&
        value != NULL)
    {
        if (self->rows != 0 &&
            self->cols != 0)
        {
            double sum = 0;
            unsigned int pos = 0;

            for(pos = 0; pos < self->rows * self->cols; pos++)
            {
                sum += self->p_arr[pos];
            }

            *value = sum / self->rows * self->cols;
        }
        else
        {
            printf("Invalid submatrix size (rows:%d,cols:%d)",
                    self->rows, self->cols);
            ret_code = JCB_EINVPTR;
        }
    }
    else
    {
        jcb_message("Invalid pointer self(%p) value(%p)", self,value);
        ret_code = JCB_EINVPTR;
    }

    return ret_code;
}


/**
 * @brief Runs a single iteration of game logic
 */
jcb_status_t
jcb_sub_board_run_step(jcb_sub_board_t *self)
{
    if (self != NULL)
    {
        if (self->p_arr     != NULL &&
            self->p_tmp_src != NULL)
        {
            int i,j;
            for( i = 0; i < self->rows; i++)
            {
                for( j = 0; j < self->cols; j++)
                {
                    double cell_value = 0;
                    jcb_status_t ret_code = __jcb_sub_board_process_cell(self,  /** Pointer to Board */
                                                                         i,     /** Y pos of cell */
                                                                         j,     /** X pos of cell */
                                                                         &cell_value); /** Cell Output value*/
            
                    if (ret_code != JCB_OK)
                    {
                        jcb_sub_board_dispose(self);
                        return ret_code;
                    } 
                    /** Change value of cell in dst game board*/
                    else
                    {
                        self->p_arr[self->cols * i + j] = cell_value;
                    }
                }
            }
        }
    }
    else
    {
        return JCB_EINVPTR;
    }
    return JCB_OK;
}


/**
 * @brief Disposes object 
 */
void
jcb_sub_board_dispose(jcb_sub_board_t *self)
{
    if (self != NULL)
    {
        if (self->p_top_boundary != NULL)
        {
            free(self->p_top_boundary);
        }

        if (self->p_bot_boundary != NULL)
        {
            free(self->p_bot_boundary);
        }

        if (self->p_arr != NULL)
        {
            free(self->p_arr);
        }

        if (self->p_tmp_src != NULL)
        {
            free(self->p_tmp_src);
        }

        free(self);
    }
}

/** ----------------------------- Private Functions --------------------------*/

/**
 * @brief Get value from temp sub matrix at given position.
 *        Depending on the boundry type of submatrix a boundry
 *        element can be returned
 *
 * @param self Pointer to current instance of Sub Board
 * @param i    Row
 * @param j    Col
 *
 * @return State of method invocation
 */
static jcb_status_t
__jcb_sub_board_get_cell(jcb_sub_board_t *self,
                       unsigned int i,
                       unsigned int j,
                       double* value)
{
    jcb_status_t ret_code = JCB_EINVPTR;

    if (self  != NULL &&
        value != NULL)
    {

        /** If Left or Right boundry is accessed */
        if ( j == -1 ||
             j >= self->cols)
        {
            *value = 0;
        }
        /** If Top or Bottom Boundry is tried to be accessed */
        else if ( self->boundary_type & JCB_TOP_BOUNDARY &&
                 (i ==  -1))
        {
            *value = self->p_top_boundary[j];
        }
        else if ( self->boundary_type & JCB_BOTTOM_BOUNDARY &&
                  (i == self->rows ))
        {
            *value = self->p_bot_boundary[j];
        }
        else
        {
            *value = self->p_tmp_src[self->cols * (i+1) + (j)];
        }
    }
    else
    {
        jcb_message("Invalid pointer self(%p) value(%p)", self,value);
        ret_code = JCB_EINVPTR;
    }

    return ret_code;
}



/**
 * @brief Performs a processing iteraton on a single cell
 *
 * @param self       Pointer to instance of Game Board
 * @param i          X-coordinate of cell
 * @param j          Y-coordinate of cell
 * @param cell (OUT) Output value
 */

static jcb_status_t
__jcb_sub_board_process_cell(jcb_sub_board_t *self,
                              unsigned int i,  
                              unsigned int j,    
                              double *cell)
{

    if (self != NULL &&
        cell != NULL)
    {
        unsigned pos = 0;

        if (self->p_arr != NULL)
        {
            double sum = 0;


            for(pos = 0; pos < 4 ; pos++)
            {
                double value = 0;
                __jcb_sub_board_get_cell(self, i+g_nbs[pos].dx, j+g_nbs[pos].dy, &value);

                sum += value;
            }
            *cell = 1.0f * sum / 4;
        }
    }
    else
    {
        jcb_message("Invalid ptr self(%p) cell(%p)\n", 
                     self,
                     cell);
        return JCB_EINVPTR;
    }

    return JCB_OK;
}




